/*
 * playitsam: Play, edit & save a TiVo video stream from stdin, file or MFS.
 * 
 * chunk.c: Chunk processing for Playitsam. $Revision: 1.33 $
 * 
 * (c) 2002, Warren Toomey wkt@tuhs.org.
 * 
 * Released under the Gnu General Public License version 2
 */

#include <sys/types.h>
#include <stdio.h>
#include <stdlib.h>
#include <sys/uio.h>
#include <unistd.h>
#ifdef __linux__
	/* David Gerdes says: Also for RedHat 7.1  I had to add this in */
# ifndef TIVO
#  include <stdint.h>
# endif
#endif
#include "playitsam.h"
#include "proto.h"
extern void Exit(int how);

#ifdef MPEG2DEC
#include "mpeg2dec/video_out.h"
#include "mpeg2dec/mpeg2.h"

extern mpeg2dec_t mpeg2dec;
extern vo_instance_t vo_out;
#endif

static FILE *cout= NULL;		/* These are used to debug the code. */
static int debug_chunk=0;		/* Normally, leave debug_chunk=0; */
static int cutmore_bframes= 0;


	/* The code in here has been quickly hacked together. */
	/* I need to a) tidy it up and b) rewrite some of it  */
	/* to make it more efficient and more elegant. Please */
	/* don't laugh too hard for now :)	- Warren      */

/*
 * Here's a commentary on the chunks and how I deal with them. The first
 * chunk in FSID streams (and from non-playitsam external sources) contains
 * junk, and so we skip these in readfile.c. So chunk 0 here is the first
 * chunk of real data.
 * 
 * Each chunk is 128K long. It starts with 4 bytes, of which the first holds the
 * number of records in the `header' of the record. I have seen 244+ records
 * in the 1st strange chunk, but normally there are less than 50 records. The
 * remaining 3 bytes of 4 I have no idea what they are for.
 * 
 * After the 4 bytes, we have the N records. Each record looks like struct
 * chunk_rec below and is 16 bytes in length. The type is either AUDIO_REC,
 * VIDEO_REC or VIDEO_REC2; however, there are some WEIRD records which we
 * skip. The size fields have to be converted into an integer using the
 * FROM_CHUNK_SIZE() macro, and converted back using the TO_CHUNK_SIZE()
 * macro.
 * 
 * Following on from the N records are the actual audio and video data. The
 * record size fields tell how big each data segment is, and they all follow
 * on contiguously. We work out the offset of each segment by summing the
 * sizes of the previous records, and adding on the sizes of the records at
 * the start, plus the 4-byte header.
 * 
 * A chunk always has one 16-byte audio segment, which seems to be a time marker
 * record, but I haven't worked out how to read it yet. A chunk has zero or
 * one 16-byte video segments, which also seems to be a time marker segment.
 * 
 * If there is a 16-byte video segment, then shortly after this there is an
 * MPEG-2 Group of Pictures (GOP) marker. This is where I cut the video
 * stream. If a chunk doesn't have a GOP marker, then I cannot place a cut
 * start or a cut end in this chunk.
 * 
 * I am assuming that, if a chunk has both 16-byte audio and video segments,
 * then these are sync points. If I cut out audio and video data from these
 * points onwards, then everything will stay in sync. This theory has yet to
 * be tested. Addendum: it seems to be false. Does anybody else have a good
 * idea how I can work out where to cut the audio to stay in sync with the
 * video.
 * 
 * I can rewrite the chunk records in such a way that stays compatible with
 * Playitsam itself, so that an cut-start or cut-end chunk still has all the
 * audio/video data, but is marked so that parts of the data will not be
 * played.
 * 
 */

/*
 * Notes on how I cut from GOP to GOP. Kees Cook (author of GOPchop, see
 * http://outflux.net/unix/software/GOPchop) pointed out that: You need to
 * remove any B-frames immediately following the I-frame of an "open" GOP.
 * The GOP header has a flag saying that it's closed, in which case you're
 * fine.
 * 
 * I then found http://www.bbc.co.uk/rd/pubs/papers/paper_14/paper_14.html
 * which says:
 * 
 * 
 * 'Bidirectionally-predictive' pictures (B-pictures) can use the previous and
 * next I- or P-pictures for motion-compensation, and offer the highest
 * degree of compression. Each block in a B-picture can be forward, backward
 * or bidirectionally predicted or intra-coded. To enable backward prediction
 * from a future frame, the coder reorders the pictures from natural
 * 'display' order to 'bitstream' order so that the B-picture is transmitted
 * after the previous and next pictures it references. This introduces a
 * reordering delay dependent on the number of consecutive B-pictures.
 * 
 * The different picture types typically occur in a repeating sequence,
 * termed a 'Group of Pictures' or GOP. A typical GOP in display order is:
 * 
 * B1 B2 I3 B4 B5 P6 B7 B8 P9 B10 B11 P12
 * 
 * The corresponding bitstream order is:
 * 
 * I3 B1 B2 P6 B4 B5 P9 B7 B8 P12 B10 B11
 */

/* How to synchronise the audio cut with the video cut: I believe that
 * there is a (non 16-octet) audio record for every MPEG-2 video frame
 * in each chunk. I am going to try this out. I will work out where the
 * first P-frame, the GOP and the P-frame after the GOP is, in relation
 * to the frames in each chunk, e.g
 *
 *               P                G                 P
 *	0     1     2     3     4     5     6    7     8     9
 *
 * When I cut from a GOP onwards, I will keep the correct # of audio
 * records before the GOP marker, e.g 4 above, and delete the rest.
 *
 * When I cut upto a GOP, I will delete the audio records before the GOP.
 *
 * This is probably all wrong, but it's the best I can do for now.
 */
 

struct chunk_rec {		/* TiVo format of a chunk record */
  u_int8_t size[3];
  u_int8_t type;
  u_int8_t unknown[12];
};

struct native_rec {		/* Native format of a chunk record */
  u_int32_t offset;		/* Offset into the chunk */
  u_int8_t type;		/* Type, given below */
};

/* The TiVo produces AUDIO_RECORD and VIDEO_RECORDs. as well as some weird
 * ones whose size is really 0. I have created two other record types to
 * mark the sections which have been removed.
 */

#define AUDIO_RECORD	0xc0
#define OLDAUDIO_RECORD	0xd0
#define VIDEO_RECORD	0xe0
#define VIDEO_RECORD2	0xf0
#define OLDVIDEO_RECORD	0xf1
#define WEIRD0x1_RECORD	0x01
#define WEIRD0x2_RECORD	0x02
#define WEIRD0x3_RECORD	0x03
#define WEIRD0x4_RECORD	0x04
#define WEIRD0x5_RECORD	0x05

#define FROM_CHUNK_SIZE(s) ((s[0] << 8 | s[1]) << 4 | (s[2] >> 4))
#define TO_CHUNK_SIZE(s,i) \
	{ s[0]= (i>>12)&0xff; s[1]= (i>>4)&0xff; s[2]= (i<<4)&0xf0; }

#define PICTURE_START 0x00 /* After a 0x0, 0x0, 0x1 marker, look for either */
#define GROUP_START   0xb8 /* video frame start, or a group of pictures start */

#define PICTURE_CODING_MASK              0x38
#define PICTURE_CODING_SHIFT             0x03
#define PICTURE_CODING_INTRA             0x01		/* I-frame marker */
#define PICTURE_CODING_PREDICTIVE        0x02		/* P-frame marker */
#define PICTURE_CODING_BIDIRECTIONAL     0x03		/* B-frame marker */


extern int verbose;
extern int allow_playback;

static int Recnum= 0;
static struct native_rec *Rec= NULL;	/* Ptr to array of records from the */
					/* last parse chunk operation */
static u_int32_t GOPoffset;		/* If non-zero, this is the position */
					/* where the GOP marker is in the buf */
static u_int32_t Boffset;		/* If non-zero, this is the position */
					/* of the 1st B-frame after the GOP */
static u_int32_t Poffset;		/* If non-zero, this is the position */
					/* of the 1st P-frame after the GOP */
static u_int32_t firstPoffset;		/* If non-zero, this is the position */
					/* of the 1st P-frame in the chunk */
static u_int32_t firstBoffset;		/* If non-zero, this is the position */
					/* of the 1st B-frame in the chunk */
static u_int8_t frameorder[30];		/* List of frame types in this chunk */
static int framecnt;			/* Never more than 30 video frames */

static u_int32_t frameoffset[30];	/* Diagnostic array, not needed */


/*
 * Read in a chunk, convert the records into native format, and find the GOP
 * offset. Return -1 if failure, 0 if no GOP or 1 if GOP. However, if findgop
 * is zero, then we don't do the scan for GOP. The latter saves CPU time.
 */
int parse_chunk(u_int8_t * buf, int chunknum, int findgop, FILE *dbfile)
{
  int i,j, pictype;
  struct chunk_rec *p;
  u_int32_t size, ofs, topofs;

  if (buf == NULL) return (-1);
  Recnum= buf[0];

  if (debug_chunk) {
    if (cout == NULL) cout= fopen("chunk.out", "w");
    dbfile= cout;
  }

  /* Allocate some storage for native details of structures */
  /* We get 1 more than needed, and store the end of data +1 in */
  /* the extra record's offset */
  if (Rec) free(Rec);
  Rec= (struct native_rec *)malloc(Recnum * (sizeof(struct native_rec) + 1));
  if (Rec == NULL) return (-1);

  /* Walk the records and find their types and sizes */
  p= (struct chunk_rec *)&buf[4];
  ofs= 4 + Recnum * sizeof(struct chunk_rec);

  for (i= 0; i < Recnum; i++, p++) {
    size= FROM_CHUNK_SIZE(p->size);
    Rec[i].type= p->type;
    Rec[i].offset= ofs;

    switch (Rec[i].type) {
      case WEIRD0x1_RECORD: continue;
      case WEIRD0x2_RECORD: continue;
      case WEIRD0x3_RECORD: continue;
      case WEIRD0x4_RECORD: continue;
      case WEIRD0x5_RECORD: continue;
      case OLDVIDEO_RECORD:
      case OLDAUDIO_RECORD:
      case AUDIO_RECORD: break;
      case VIDEO_RECORD2:
      case VIDEO_RECORD: break;
      default:
#if 0
	fprintf(stderr,
	"\r\nPlease tell Warren wkt@tuhs.org about the 0x%x records\r\n",
							Rec[i].type);
#endif
       	continue;
    }
    ofs += size;
  }
  Rec[i].offset= ofs;		/* Set up the fake offset after last record */
  Rec[i].type= WEIRD0x1_RECORD;
  if (findgop==0) return(0);

  /*
   * Walk the video record to search for:
   *	a GOP marker,
   *	a B-frame after a GOP marker,
   *	a P-frame after a GOP marker,
   *	the first B-frame,
   *	the first P-frame.
   * Also make a list of which video frames are in this chunk.
   */
  GOPoffset= Boffset= Poffset= firstPoffset= firstBoffset= 0; framecnt=0;

  for (i= 0; i < Recnum; i++) {
    if ((Rec[i].type!= VIDEO_RECORD) && (Rec[i].type!= VIDEO_RECORD2)) continue;

    /* Find the top offset in this group of contiguous video records */
    for (j= i+1; j <= Recnum; j++) {
      if ((Rec[j].type!= VIDEO_RECORD) && (Rec[j].type!= VIDEO_RECORD2)) {
        topofs= Rec[j].offset; break;
      }
    }

    /* Now we know the real extent of this group of V records, search for
     * GOPs, B-, I- and P- frames.
     */
    for (ofs= Rec[i].offset; ofs< topofs-4; ofs++) {
      if (buf[ofs] != 0x0) continue;
      if (buf[ofs+1] != 0x0) continue;
      if (buf[ofs+2] != 0x1) continue;		/* Ignore non-marker stuff */

      switch (buf[ofs+3]) {
        case GROUP_START: if (GOPoffset == 0) GOPoffset= ofs; break;

        case PICTURE_START:
          pictype= (buf[ofs + 5] & PICTURE_CODING_MASK) >> PICTURE_CODING_SHIFT;

          switch(pictype) {
	    case PICTURE_CODING_INTRA:
	      frameoffset[framecnt]= ofs;
	      frameorder[framecnt++]= PICTURE_CODING_INTRA; break;

	    case PICTURE_CODING_BIDIRECTIONAL:
	      frameoffset[framecnt]= ofs;
	      frameorder[framecnt++]= PICTURE_CODING_BIDIRECTIONAL;
	      if (firstBoffset == 0) firstBoffset= ofs;
	      if ((Boffset == 0) && (GOPoffset > 0) && (ofs > GOPoffset)) Boffset= ofs;
	      break;

	    case PICTURE_CODING_PREDICTIVE:
	      frameoffset[framecnt]= ofs;
	      frameorder[framecnt++]= PICTURE_CODING_PREDICTIVE;
	      if (firstPoffset == 0) firstPoffset= ofs;
	      if ((Poffset == 0) && (GOPoffset > 0) && (ofs > GOPoffset)) Poffset= ofs;
  	      break;
  	  }
       }
    }
    i=j;	/* We do this to force the outside loop to skip */
		/* all the V records that we walked */
		/* through in the preceding loop */
  }

  if (dbfile!=NULL) {
    fprintf(dbfile, "\nChunk %d\n", chunknum);
    for (i=0;i<framecnt;i++) {
      switch(frameorder[i]) {
	case PICTURE_CODING_INTRA: fprintf(dbfile, "I %d\n",frameoffset[i]); break;
	case PICTURE_CODING_PREDICTIVE: fprintf(dbfile, "P %d\n",frameoffset[i]); break;
	case PICTURE_CODING_BIDIRECTIONAL: fprintf(dbfile, "B %d\n",frameoffset[i]);break;
      }
    }
    if (GOPoffset) fprintf(dbfile, "Found GOP at %d\n", GOPoffset);
    if (firstPoffset) fprintf(dbfile, "Found 1st P-frame at %d\n", firstPoffset);
    if (firstBoffset) fprintf(dbfile, "Found 1st B-frame at %d\n", firstBoffset);
    if (Boffset) fprintf(dbfile, "Found B-frame after GOP at %d\n", Boffset);
    if (Poffset) fprintf(dbfile, "Found P-frame after GOP at %d\n", Poffset);
    fflush(dbfile);
  }
  if (GOPoffset) return (1); else return (0);
}


/*
 * This is purely a diagnostic routine to print out the TiVo records. This is
 * to verify that the cut routines below are actually doing the correct
 * thing.
 */
void printout_recs(u_int8_t * buf, FILE *dbfile)
{
  int i, recnum;
  struct chunk_rec *p;
  u_int32_t size;

  if (dbfile==NULL) return;
  if (buf == NULL) return;

  recnum= buf[0];
  fprintf(dbfile, "# records: %d\n", recnum);

  /* Walk the records and find their types and sizes */
  p= (struct chunk_rec *)&buf[4];

  for (i= 0; i < recnum; i++, p++) {
    size= FROM_CHUNK_SIZE(p->size);
    fprintf(dbfile, "    Rec %2d: ",i);
    switch(Rec[i].type) {
      case AUDIO_RECORD: fprintf(dbfile, " A "); break;
      case OLDAUDIO_RECORD: fprintf(dbfile, "XA "); break;
      case VIDEO_RECORD2:
      case VIDEO_RECORD: fprintf(dbfile, " V "); break;
      case OLDVIDEO_RECORD: fprintf(dbfile, "XV "); break;
      default: fprintf(dbfile, " ? "); break;
    }
    fprintf(dbfile, "size %05d offset %05d", size, Rec[i].offset);
    if (Rec[i].offset==GOPoffset) fprintf(dbfile, " GOP");
    if (Rec[i].offset==firstPoffset) fprintf(dbfile, " 1st P");
    if (Rec[i].offset==firstBoffset) fprintf(dbfile, " 1st B");
    if (Rec[i].offset==Boffset) fprintf(dbfile, " B after GOP");
    if (Rec[i].offset==Poffset) fprintf(dbfile, " P after GOP");
    fprintf(dbfile, "\n");
  }
  fprintf(dbfile, "Done\n"); fflush(dbfile); return;
}


/*
 * Given our own struct native_rec Rec[] array, convert back to the TiVo
 * format and update the chunk. We must do this every time we modify the
 * Rec[] array.
 */
void update_tivorecs(u_int8_t * buf)
{
  int i, recnum;
  struct chunk_rec *p;
  u_int32_t size;

  if (buf == NULL) return;
  recnum= buf[0];
  p= (struct chunk_rec *)&buf[4];

  for (i= 0; i < recnum; i++, p++) {
    size= Rec[i+1].offset - Rec[i].offset;
    TO_CHUNK_SIZE(p->size, size); p->type= Rec[i].type;
  }
  if (debug_chunk) printout_recs(buf, cout);
}


/*
 * Given a buffer, and the assumption that we have already run parse_chunk on
 * the buffer to create the struct native_rec *Rec, play (demultiplex) the
 * chunk to the given audio and video file descriptors.
 */
void play_chunk(u_int8_t * buf, int audio_fd, int video_fd)
{
  int i;
  u_int32_t size;

  for (i= 0; i < Recnum; i++) {
    size= Rec[i+1].offset - Rec[i].offset;
    switch (Rec[i].type) {
    case VIDEO_RECORD2:
    case VIDEO_RECORD:
#ifndef TIVO
      if (size != 0x10)
#endif
#ifdef MPEG2DEC
        mpeg2_decode_data(&mpeg2dec, (uint8_t*)&buf[Rec[i].offset],
			((uint8_t *) (&buf[Rec[i].offset])) + size);
#else
	write(video_fd, &buf[Rec[i].offset], size);
#endif
      break;
    case AUDIO_RECORD:
#ifndef TIVO
      if (size != 0x10)
#endif
	write(audio_fd, &buf[Rec[i].offset], size);
      break;
    }
  }
}


/*
 * The format of a GOP marker I have gleaned from GOPchop, see
 * http://outflux.net/unix/software/GOPchop. I believe that the marker starts
 * with the 4 octets 0x0, 0x0, 0x1, 0xB8. This is then followed by four more
 * octets:
 * 
 *         4         5         6        7
 *               |         |        |         |
 *      7 65432 107654 3 210765 432107 6 543210
 *      1 11111 111111 1 111111 111111 1 1
 *      d hour  min    m sec    pic    c b
 *      r              a               l roken
 *      o              r               osed
 *      p              k
 *
 * 
 * GOPchop sets both the broken and closed flags, so I do the same below.
 */
#define GOP_FLAG_BROKEN  0x20
#define GOP_FLAG_CLOSED  0x40


/*
 * Remove all of the video data starting from the GOP marker onwards.
 * To remove the video data, we find the video record which contains the
 * first octet of the GOP marker, and reduce its offset. Then we mark all
 * the remaining records as non-video.
 *
 * For the audio, we keep the appropriate # of non-16 audio records that
 * occurred before the GOP, and mark the others as non-audio.
 */
void cut_chunk_GOPonwards(u_int8_t * buf)
{
  int i, keepaudioframe=0;
  u_int32_t vend, vstart;
  struct chunk_rec *p;

  if (GOPoffset == 0) return;			/* Sanity checks */
  if (buf == NULL) return;
  if (debug_chunk) { fprintf(cout, "Cut_chunk_GOPonwards()\n"); fflush(cout); }

  /* Work out how many audio records to keep: count the # of frames before
   * the I-frame.
   */
  for (i=0;(i<framecnt) && (frameorder[i]!=PICTURE_CODING_INTRA);i++) keepaudioframe++;

  p= (struct chunk_rec *)&buf[4];
  for (i= 0; i < Recnum; i++, p++) {
    switch (Rec[i].type) {
    case AUDIO_RECORD:
      if ((Rec[i+1].offset-Rec[i].offset)==16) break;
      if (keepaudioframe-- >0) break;
      Rec[i].type= OLDAUDIO_RECORD; break;

    case VIDEO_RECORD2:
    case VIDEO_RECORD:
      vstart= Rec[i].offset;
      vend= Rec[i+1].offset - 1;

      if (vstart > GOPoffset) { Rec[i].type= OLDVIDEO_RECORD; break; }

      if ((vstart <= GOPoffset) && (vend >= GOPoffset)) {
	Rec[i].offset= GOPoffset; Rec[i].type= OLDVIDEO_RECORD;
	break;
      }
    }
  }
  update_tivorecs(buf);
}


/*
 * This is a very tricky routine. This chunk is the end of a cut. We must
 * `remove' all of the video data up to but not including the GOP marker. We
 * mark all the video data before the GOP marker as invalid. For the segment
 * containing the GOP marker, we have to reset its offset.
 * 
 * Once this is done, we need to find any B-frames after the GOP marker but
 * before a P-frame, and `remove' them as well. We need to use the same
 * technique as in cut_chunk_GOPonwards() to cut out the B-frame, and the
 * same technique with the GOP marker here to restart the data with the
 * beginning of the P-frame.
 * 
 * Returns 0 if this is the end of the B-frame cutting, or 1 if we need to also
 * work on the next chunk as well. If we return 2, we haven't found a B-frame
 * after the GOP here, so search for it in the cut_morebf() routine.
 */
int cut_chunk_uptoGOP(u_int8_t * buf)
{
  int i, tossaudioframe, Poffset_waszero;
  u_int32_t vend, vstart;
  struct chunk_rec *p;

  if (GOPoffset == 0) return (0);			/* Sanity checks */
  if (buf == NULL) return (0);
  if (debug_chunk) { fprintf(cout, "Cut_chunk_uptoGOP()\n"); fflush(cout); }

  /* Work out how many audio records to discard: count the # of frames before
   * the I-frame. XXXXXX WARNING XXXXXX Experimental Section. I add 2 here to
   * try & toss out the audio for the 2 B frames. I have no idea if this is
   * the right thing to do.
   */
#if 0
  tossaudioframe=2;
  for (i=0;(i<framecnt) && (frameorder[i]!=PICTURE_CODING_INTRA);i++) tossaudioframe++;
#else
  tossaudioframe=0;
#endif

  /* Mark the GOP as closed and broken */
  buf[GOPoffset + 7] |= (GOP_FLAG_BROKEN | GOP_FLAG_CLOSED);

  /* Do the `removing' of video data before the GOP first */
  p= (struct chunk_rec *)&buf[4];
  for (i= 0; i < Recnum; i++, p++) {
    if (Rec[i].type == AUDIO_RECORD) {
      if ((Rec[i+1].offset-Rec[i].offset)==16) continue;
      if (tossaudioframe-- >0) Rec[i].type= OLDAUDIO_RECORD;
      continue;
    }

    vstart= Rec[i].offset;
    vend= Rec[i+1].offset - 1;

    /* Keep any 16-byte video record */
    if ((vend - vstart)==15) continue;

    if (GOPoffset > vend) { Rec[i].type= OLDVIDEO_RECORD; continue; }

    if ((vstart <= GOPoffset) && (vend >= GOPoffset)) {
      Rec[i].offset= GOPoffset; continue; break;
    }
  }

  /* Sometimes there is no  B-frame after the GOP */
  /* This means that we need to find the B-frame */
  /* in the next frame, can cut from there to the P-frame. */
  if (Boffset == 0) { update_tivorecs(buf); return (2); }

#define CHUNK_SIZE (256*512)

  Poffset_waszero= 0;
  					/* If no Poffset, then set it to be above */
  					/* the end of the last video record */
  					/* as it must be in next chunk */
  if (Poffset == 0) { Poffset_waszero= 1; Poffset= CHUNK_SIZE; }

  /* Now deal with the removal of data between the B-frames and the P-frame */
  p= (struct chunk_rec *)&buf[4];
  for (i= 0; i < Recnum; i++, p++) {
    if ((Rec[i].type != VIDEO_RECORD)&&(Rec[i].type != VIDEO_RECORD2)) continue;

    vstart= Rec[i].offset;
    vend= Rec[i+1].offset - 1;

    /* Keep any 16-byte video record */
    if ((vend - vstart)==15) continue;

    				/* Skip video records which fall before */
    				/* the B-frame or after the P-frame */
    if ((vend < Boffset) || (vstart > Poffset)) continue;

    				/* If the record lies between the Boffset */
				/* and the Poffset, simply mark it as an */
				/* invalid record */
    if ((vstart > Boffset) && (vend < Poffset)) {
      Rec[i].type= OLDVIDEO_RECORD; continue;
    }

				/* If the Boffset lies within this record, */
				/* shift the record offset to Boffset, and */
				/* mark as invalid. */
    if ((vstart <= Boffset) && (vend >= Boffset)) {
      Rec[i].type= OLDVIDEO_RECORD; Rec[i].offset= Boffset; continue;
    }

				/* If the Poffset lies within this record, */
				/* shift the record offset to Poffset */
    if ((vstart <= Poffset) && (vend >= Poffset)) {
      Rec[i].offset= Poffset; continue;
    }
  }
  update_tivorecs(buf); return (Poffset_waszero);
}

/*
 * This routine gets called for the buffer immediately after the one which
 * was edited by the cut_chunk_uptoGOP() routine, and only when there was no
 * P-frame after the GOP in that chunk. Here we trim out the initial records
 * up to the first P-frame.
 *
 * If lookforB is 2, we don't cut before the first B frame.
 */
void cut_morebf(u_int8_t * buf, int lookforB)
{
  int i;
  u_int32_t vend, vstart;
  struct chunk_rec *p;

  if (buf == NULL) return;			/* Sanity checks */
  if (firstPoffset == 0) {
    if (debug_chunk) { fprintf(cout, "AARGH in cut_morebf()\n"); fflush(cout); }
    return;
  }
  if (debug_chunk) { fprintf(cout, "Cut_morebf()\n"); fflush(cout); }

  /* Set firstBoffset to 0 if we are not looking for the first B-frame */
  if (lookforB!=2) firstBoffset=0;

  /* Do the `removing' of video data before the P-frame
   * and after the B-frame inclusive.
   */
  p= (struct chunk_rec *)&buf[4];
  for (i= 0; i < Recnum; i++, p++) {
    if ((Rec[i].type != VIDEO_RECORD)&&(Rec[i].type != VIDEO_RECORD2)) continue;
      vstart= Rec[i].offset;
      vend= Rec[i+1].offset - 1;

      /* Keep any 16-byte video record */
      if ((vend - vstart)==15) continue;

      if (vstart> firstPoffset) continue;
      if (vend< firstBoffset) continue;

      if ((vstart > firstBoffset) && (vend < firstPoffset)) {
	Rec[i].type= OLDVIDEO_RECORD; continue;
      }

      if ((vstart <= firstBoffset) && (vend >= firstBoffset)) {
	Rec[i].offset= firstBoffset; Rec[i].type= OLDVIDEO_RECORD; continue;
      }

      if ((vstart <= firstPoffset) && (vend >= firstPoffset)) {
	Rec[i].offset= firstPoffset; continue;
      }
  }
  update_tivorecs(buf);
}

/* This routine flips all the video and audio records to be
 * old video & audio records, and vice versa.
 */
void flip_record_types(u_int8_t *buf)
{
  int i;
  struct chunk_rec *p;

  if (cout) fprintf(cout, "\nIn flip\n");

  p= (struct chunk_rec *)&buf[4];
  for (i= 0; i < Recnum; i++, p++) {
    switch(Rec[i].type) {
      case VIDEO_RECORD2:
      case VIDEO_RECORD: Rec[i].type=OLDVIDEO_RECORD; break;
      case AUDIO_RECORD: Rec[i].type=OLDAUDIO_RECORD; break;
      case OLDVIDEO_RECORD: Rec[i].type=VIDEO_RECORD; break;
      case OLDAUDIO_RECORD: Rec[i].type=AUDIO_RECORD; break;
    }
  }
  printout_recs(buf, cout);
}

/* This is the nub of the editing process during playback/record/edit.
 * We are given a buf to play, and possibly an old buf. If the old buf
 * is non-NULL, then we got up to the GOP on the previous use of
 * edit_and_play_chunk(), and so we have to play from the GOP onwards.
 *
 * If the variable stop_at_gop is true, we play up to the GOP. If the
 * variable skip_cuts is non-zero, then we don't call
 * find_chunk_in_cutlist() to work out what to do with this chunk.
 *
 * We return -1 if something went wrong, 0 if everything is ok, and 1
 * if we need another frame straightaway. The latter includes needing
 * another chunk to find and kill the B-frames, or if there was no GOP
 * to stop at when stop_at_gop is true.
 *
 * If vid_fd ==0, we are actually exporting a raw tystream to aud_fd
 * and not playing the stream. So we don't always have to find a GOP.
 */

static u_int8_t *oldbuf=NULL;
static int lastchunk=-1;		/* Chunk before the one we're up to */

/* Need to do this before recording */
void reset_lastchunk(void)
{ lastchunk=-1; oldbuf=NULL; }

int edit_and_play_chunk(u_int8_t *buf, int chk, int aud_fd, int vid_fd,			 int stop_at_gop, int skip_cuts)
{
  int where_is_chunk;
  int numleft, numsent;
  unsigned char *bptr;
  int found_a_gop=0;
  int is_a_jump=0;

  if (cout!=NULL) fprintf(cout,
	"EAPC, chunk %d lbuf %ux stop %d skip %d\n", chk, (int) oldbuf,
						stop_at_gop, skip_cuts);

  /* First work out if this is a jump. If so, we don't use oldbuf, and
   * we also can't stop at the GOP in this chunk.
   */
  if ((chk-lastchunk)!=1) is_a_jump=1;
  if (cout!=NULL) fprintf(cout, "is_a_jump is %d\n", is_a_jump);
  if (is_a_jump) { oldbuf=NULL; stop_at_gop=0; }
  lastchunk=chk;

  /* Oldbuf is non-NULL. Assume that the Rec[] array still holds the
   * information about this chunk. Flip the audio & video records, and
   * play it.
   */
  if (oldbuf) { flip_record_types(oldbuf); play_chunk(oldbuf, aud_fd, vid_fd); }

    				/* Work out if we should play this chunk */
  if (skip_cuts) where_is_chunk= find_chunk_in_cutlist(chk);
  else where_is_chunk= IS_OUTSIDE_CUT;
  if (is_a_jump) where_is_chunk= IS_CUT_END;
  if (cutmore_bframes) where_is_chunk= IS_WANTPFRAME;
  				/* Stop at the GOP if we have been asked to. */
  if (stop_at_gop && (where_is_chunk==IS_OUTSIDE_CUT))
    where_is_chunk=IS_CUT_START;

  /* Special case when we are exporting a tystream and not
   * playing it. Only have to find a GOP on the edge of a cut.
   */
  if (vid_fd==0) {
    switch(where_is_chunk) {
      case IS_INSIDE_CUT:  return(0);
      case IS_OUTSIDE_CUT: break;
      case IS_CUT_START:   parse_chunk(buf, chk, 1, NULL);
			   cut_chunk_GOPonwards(buf); break;
      case IS_CUT_END:     parse_chunk(buf, chk, 1, NULL);
			   cutmore_bframes= cut_chunk_uptoGOP(buf); break;
      case IS_WANTPFRAME:  parse_chunk(buf, chk, 1, NULL);
			   cut_morebf(buf, cutmore_bframes);
			   cutmore_bframes= 0; break;
    }
    /* Save the chunk */
    numleft= CHUNK_SIZE; bptr= buf;
    while (numleft > 0) {
      numsent= write(aud_fd, bptr, numleft);
      if (numsent < 1) { fprintf(stderr, "Failure saving chunk\n"); Exit(1); }
      numleft -= numsent; bptr += numsent;
    }
    return(0);
  }

  /* Normal edit/playback mode now. */
  if (where_is_chunk==IS_INSIDE_CUT) return(1);
  found_a_gop=parse_chunk(buf, chk, 1, NULL);

  if (cout!=NULL) fprintf(cout, "Just before switch, whereis= %d\n",
			where_is_chunk); fflush(cout);

  switch(where_is_chunk) {
    case IS_OUTSIDE_CUT: break;
    case IS_CUT_START:   if (found_a_gop) cut_chunk_GOPonwards(buf); break;
    case IS_CUT_END:     if (found_a_gop)
			    cutmore_bframes= cut_chunk_uptoGOP(buf);
		         break;
    case IS_WANTPFRAME:  cut_morebf(buf, cutmore_bframes);
			 cutmore_bframes= 0; break;
  }
  if (allow_playback) play_chunk(buf, aud_fd, vid_fd);

  if (stop_at_gop && found_a_gop) oldbuf=buf; else oldbuf=NULL;
  if (cutmore_bframes) return(1);
  if (is_a_jump) return(1);
  if (!found_a_gop) return(1);
  return(0);
}
