/*
 * playitsam: Play, edit & save a TiVo video stream from stdin, file or MFS.
 *
 * typarse.c: Diagnostic parser of tystreams. $Revision: 1.18 $
 *
 * (c) 2002, Warren Toomey wkt@tuhs.org. 
 *
 * Released under the Gnu General Public License version 2
 */

#include <sys/types.h>
#include <stdio.h>
#include <signal.h>
#include <fcntl.h>
#include <unistd.h>
#define CHUNK_SIZE 256*512

#define PICTURE_START	0x00
#define GROUP_START	0xb8
#define MPEG_END	0xb9
#define PACK_START	0xba
#define SYSTEM_START	0xbb

#define PICTURE_CODING_MASK              0x38
#define PICTURE_CODING_SHIFT             0x03
#define PICTURE_CODING_INTRA             0x01
#define PICTURE_CODING_PREDICTIVE        0x02
#define PICTURE_CODING_BIDIRECTIONAL     0x03

/* The TiVo produces AUDIO_RECORD and VIDEO_RECORDs. as well as some weird
 * ones whose size is really 0. I have created two other record types to
 * mark the sections which have been removed.
 */

#define AUDIO_RECORD    0xc0
#define OLDAUDIO_RECORD 0xd0
#define VIDEO_RECORD    0xe0
#define VIDEO_RECORD2   0xf0
#define OLDVIDEO_RECORD 0xf1
#define WEIRD0x1_RECORD 0x01
#define WEIRD0x2_RECORD 0x02
#define WEIRD0x3_RECORD 0x03
#define WEIRD0x4_RECORD 0x04
#define WEIRD0x5_RECORD 0x05

int verbose=0;

int find_start_code(uint8_t * source)
{
	if ((source[0]==0x0) && (source[1]==0x0) && (source[2]==0x1))
	  return(source[3]);
        return -1;
}

/* Find offset of any GOP in a video chunk, and print it out */
int find_GOP(uint8_t *buf, size_t num,int ofs)
{
 size_t i, total;
 int res, pictype;

 total= num - 4;

 for (i=0;i<total;i++,buf++,ofs++) {
  res= find_start_code(buf);
  switch(res) {
    case PICTURE_START: printf("Found picture start at %d", ofs);
	pictype= (buf[5] & PICTURE_CODING_MASK) >> PICTURE_CODING_SHIFT;
	switch (pictype) {
	  case PICTURE_CODING_INTRA: printf(", I-frame\n"); break;
	  case PICTURE_CODING_PREDICTIVE: printf(", P-frame\n"); break;
	  case PICTURE_CODING_BIDIRECTIONAL: printf(", B-frame\n"); break;
	  default: printf(", Unknown-frame\n"); break;
	}
	break;
    case GROUP_START: printf("Found GOP at %d\n", ofs); break;
    case MPEG_END: printf("Found MPEG end at %d\n", ofs); break;
    case PACK_START: printf("Found pack start at %d\n", ofs); break;
    case SYSTEM_START: printf("Found system start at %d\n", ofs); break;
  }
 }
 return(0);
}

void print_time(char *str, uint8_t *cptr, int offset)
{
 unsigned long tim;
 double seconds;
#define ONE_SECOND 90000.0

 if (cptr[offset+8]!=0x07) return;
 if (cptr[offset+9]!=0x21) return;
 tim=((((((((cptr[offset+9]&0x06)<<7)|cptr[offset+10])<<7)|
     (cptr[offset+11]>>1))<<8)|cptr[offset+12])<<7)|(cptr[offset+13]>>1);

 seconds = ((double)tim) / ONE_SECOND;
 printf("%s time is %lu or %f seconds\n",str, tim, seconds);
}

void print_somehex(uint8_t *cptr, int offset, int num)
{
 int i;
 int total;

 if ((offset+num)>CHUNK_SIZE) { printf("ERR"); return; }

 total=16; if (num<total) total=num;

 for (i=0;i<total;i++) {
   printf("%02x ",(cptr[offset+i] & 0xff));
 }
}

static void
parse_chunk(uint8_t buf[CHUNK_SIZE], int chunknum)
{
  uint8_t type, rec_type[256];
  unsigned int ofs, rec_start[256];
  unsigned int size, rec_size[256];
  uint8_t num_recs = buf[0];
  int i;
  uint8_t *p;
  int total_audio = 0;
  int total_video = 0;
  int numaud = 0;
  int numvid = 0;
  int first_vidoffset=0;
  int overall_total;
  int unknown_recs = 0;
  int vid16=0;
  int aud16=0;
  int seen_an_error=0;

  printf("Chunk %d has %d records\n",chunknum, num_recs);
  /*
   * each 128k chunk starts with N 16 byte records that tell you what sorts
   * of things are in the chunk
   */
  p = &buf[4];
  ofs = 4 + num_recs * 16;

  for (i = 0; i < num_recs; i++) {
    rec_start[i]=ofs;
    size = (p[0] << 8 | p[1]) << 4 | (p[2] >> 4); rec_size[i]=size;
    type = p[3]; rec_type[i]=type;

    printf(" %02d sz %05d off %06d", i, size, ofs);

    switch(type) {
				/* Don't add size on for weird records */
      case WEIRD0x1_RECORD: printf(" 0x1\n");  p += 16; continue;
      case WEIRD0x2_RECORD: printf(" 0x2\n");  p += 16; continue;
      case WEIRD0x3_RECORD: printf(" 0x3\n");  p += 16; continue;
      case WEIRD0x4_RECORD: printf(" 0x4\n");  p += 16; continue;
      case WEIRD0x5_RECORD: printf(" 0x5\n");  p += 16; continue;

      case VIDEO_RECORD:
        printf("  V: ");
	goto here;
      case VIDEO_RECORD2:
        printf(" V2: ");
here:
	numvid++;
        if (first_vidoffset==0) first_vidoffset=ofs;
	if ((size==16) && (vid16==0)) vid16=ofs;
	print_somehex(buf,ofs,size);
        printf("\n");
	/* if (size==16) print_time("V", buf,ofs); */
        total_video += size;
	break;
      case AUDIO_RECORD:
	numaud++;
        printf("  A: ");
	if ((size==16) && (aud16==0)) aud16=ofs;
        print_somehex(buf,ofs,size);
        printf("\n");
	/* if (size==16) print_time("A", buf,ofs); */
        total_audio += size;
	break;
      case OLDAUDIO_RECORD:
        printf(" OA: ");
        print_somehex(buf,ofs,size);
        printf("\n");
	break;
      case OLDVIDEO_RECORD:
        printf(" OV: ");
        print_somehex(buf,ofs,size);
        printf("\n");
	break;
      default:
        printf(" unknown (0x%02x)\n",type);
        unknown_recs++;
    }
    if (ofs>=CHUNK_SIZE) {
      if (!seen_an_error) printf("ERROR: records to here add up to > 128K!\n");
      seen_an_error=1;
    }

    p += 16; ofs+= size;
  }

  if (aud16) printf("      aud16 at %d\n",aud16);
  if (vid16) printf("      vid16 at %d\n",vid16);
  printf("Found %d audio records\n",numaud);
  printf("Found %d video records\n",numvid);

  if ((seen_an_error==0) && (first_vidoffset!=0)) {
    ofs= find_GOP(&buf[first_vidoffset], total_video, first_vidoffset);
    if (ofs && vid16) printf("GOP - vid16 is %d\n", ofs - vid16);
    for (i = 0; i < num_recs; i++) {
	if ((rec_size[i]==16) && (ofs>=rec_start[i]) &&
		(ofs < (rec_start[i] + rec_size[i])) )
		printf("Naughty GOP in a 16-byte record\n");
    }
  }

  overall_total= total_audio + total_video + num_recs * 16 + 4;
  printf("  total video: 0x%x total audio: 0x%x total: 0x%x\n",
        total_video, total_audio, overall_total);
  printf("  total video: %d total audio: %d total: %d diff %d\n\n",
        total_video, total_audio, overall_total, CHUNK_SIZE - overall_total);
}

static void
usage(void)
{
  printf("typarse [options] <infile>\n");
  printf("options:\n");
  printf("  -v                 verbose\n");
  printf("  -h                 help\n");
  exit(1);
}

int
main(int argc, char *argv[])
{
  char *fname;
  int fd, chunk=0;
  uint8_t buf[CHUNK_SIZE];
  extern char *optarg;
  extern int optind;
  int c;

  while ((c = getopt(argc, argv, "vh")) != -1) {
    switch (c) {
    case 'v': verbose++; break;
    case 'h':
    default: usage();
    }
  }

  argc -= optind;
  argv += optind;

  if (argc < 1) usage();
  fname = argv[0];
  fd = open(fname, O_RDONLY);

  /* Read the first chunk. If it doesn't contain F5 46 XX BD, then it's
   * a valid chunk, else skip it.
   */
  read(fd, buf, CHUNK_SIZE);
  if ((buf[0]!=0xf5) && (buf[1]!=0x46) && (buf[3]!=0xbd)) {
    parse_chunk(buf,0); chunk=1;
  }
  else printf("Skipped useless chunk 0\n");

  while (read(fd, buf, CHUNK_SIZE) == CHUNK_SIZE) {
    parse_chunk(buf,chunk); chunk++;
  }
  return(0);
}
