/*
 * playitsam: Play, edit & save a TiVo video stream from stdin, file or MFS.
 *
 * vsplit.c: Split tystreams into MPEG audio & video. $Revision: 1.14 $
 *
 * (c) 2002, Warren Toomey wkt@tuhs.org.
 *
 * Released under the Gnu General Public License version 2
 */

#include <sys/types.h>
#include <stdio.h>
#include <fcntl.h>
#include <unistd.h>
#define CHUNK_SIZE 256*512

/* The TiVo produces AUDIO_RECORD and VIDEO_RECORDs. as well as some weird
 * ones whose size is really 0. I have created two other record types to
 * mark the sections which have been removed.
 */

#define AUDIO_RECORD    0xc0
#define OLDAUDIO_RECORD 0xd0
#define VIDEO_RECORD    0xe0
#define VIDEO_RECORD2   0xf0
#define OLDVIDEO_RECORD 0xf1
#define WEIRD0x1_RECORD 0x01
#define WEIRD0x2_RECORD 0x02
#define WEIRD0x3_RECORD 0x03
#define WEIRD0x4_RECORD 0x04
#define WEIRD0x5_RECORD 0x05

FILE *audio_file;
FILE *video_file;
int verbose = 0;
unsigned long highest_total=0, highest_count=0;
unsigned long lowest_total=0,  lowest_count=0;

/* Return the timestamp in a 16-byte A or V
 * record, in milliseconds. 0 is returned if
 * error.
 */
unsigned long get_timestamp(uint8_t *cptr)
{
 unsigned long tim, milli;
#define ONE_MILLISECOND 90

 if (cptr[8]!=0x07) return(0);
 if (cptr[9]!=0x21) return(0);
 tim=((((((((cptr[9]&0x06)<<7)|cptr[10])<<7)|
     (cptr[11]>>1))<<8)|cptr[12])<<7)|(cptr[13]>>1);

 milli = tim / ONE_MILLISECOND;
 return(milli);
}

void parse_chunk(unsigned char buf[CHUNK_SIZE])
{
  int num_recs = buf[0];
  int i;
  unsigned char *p;
  int total_audio = 0;
  int total_video = 0;
  int ofs = 0;
  int unknown_recs = 0;
  long aposn;
  unsigned int audsize;
  unsigned long lowest_atime=0, lowest_vtime=0;
  unsigned long highest_atime=0, highest_vtime=0;
  unsigned long tim;

  /*
   * each 128k chunk starts with N 16 byte records that tell you what sorts
   * of things are in the chunk
   */
  p = &buf[4];
  ofs = 4 + num_recs * 16;

  for (i = 0; i < num_recs; i++) {
    unsigned size = (p[0] << 8 | p[1]) << 4 | (p[2] >> 4);
    unsigned type = p[3];

    if (type < 0x10) { p += 16; continue; }

    if ((type == VIDEO_RECORD)||(type == VIDEO_RECORD2)) {
      if (size != 0x10) {
	fwrite(&buf[ofs], 1, size, video_file);
	if (verbose) { printf("video at rec %d ofs 0x%x size 0x%x\n",
							i, ofs, size); }
      } else {
	tim= get_timestamp(&buf[ofs]);
	if (lowest_vtime==0) lowest_vtime=tim;
	if (highest_vtime==0) highest_vtime=tim;
	if (tim<lowest_vtime) lowest_vtime=tim;
	if (tim>highest_vtime) highest_vtime=tim;
      }
      total_video += size;
      ofs += size;
      p += 16;
      continue;
    }
    if (type == AUDIO_RECORD) {
      audsize= size;
      if ((size != 0x10) || ( (size == 0x10) && (buf[ofs]!=0x00))) {

        if (size==867) { audsize= 864; printf("Ha hit a size 0x867 A rec\n"); }
	fwrite(&buf[ofs], 1, audsize, audio_file);
	if (verbose) { printf("audio at rec %d ofs 0x%x size 0x%x\n",
							i, ofs, size); }
      }
      if ((size == 0x10) && (buf[ofs]!=0x00)) {
	aposn=ftell(audio_file);
	printf("Ha just hit a size 10 and !00 at 0x%lx\n",aposn);
      }
      if (size == 0x10) {
	tim= get_timestamp(&buf[ofs]);
	if (lowest_atime==0) lowest_atime=tim;
	if (highest_atime==0) highest_atime=tim;
	if (tim<lowest_atime) lowest_atime=tim;
	if (tim>highest_atime) highest_atime=tim;
      }
      total_audio += size;
      ofs += size;
      p += 16;
      continue;
    }
    unknown_recs++;
    p += 16;
    if (type == 0x01) continue;
    if (type == 0x02) continue;
    ofs+= size;
  }

  if (verbose) 
    printf("recs: %d unk: %d vid: 0x%x aud: 0x%x total: 0x%x ofs=0x%x\n",
	   num_recs, unknown_recs, total_video, total_audio,
	   total_audio + total_video + num_recs * 16 + 4, ofs);

  /* Save the difference in v and a times so we can do an average later */
  tim= highest_vtime - highest_atime;
  if (tim < 2000) { highest_total+= tim; highest_count++; }
  tim= lowest_vtime - lowest_atime;
  if (tim < 2000) { lowest_total+= tim; lowest_count++; }
}

void usage(void)
{
  printf("vsplit [options] <infile> <audiofile> <videofile>\n");
  printf("options:\n");
  printf("    -v  verbose\n");
  printf("    -h  help\n");
  exit(1);
}

int main(int argc, char *argv[])
{
  char *fname;
  FILE *zin;
  unsigned char buf[CHUNK_SIZE];
  extern char *optarg;
  extern int optind;
  int c;
  unsigned long average;

  while ((c = getopt(argc, argv, "vh")) != -1) {
    switch (c) {
    case 'v': verbose++; break;
    case 'h':
    default: usage(); exit(1);
    }
  }

  argc -= optind; argv += optind;

  if (argc < 3) { usage(); }

  fname = argv[0];
  zin = fopen(fname, "r");
  if (zin == NULL) { printf("Cannot read file %s\n", fname); exit(1); }
  audio_file = fopen(argv[1], "w");
  if (audio_file== NULL) { printf("Cannot write file %s\n", argv[2]); exit(1); }
  video_file = fopen(argv[2], "w");
  if (video_file== NULL) { printf("Cannot write file %s\n", argv[1]); exit(1); }

  /* Read the first chunk. If it doesn't contain F5 46 XX BD, then it's
   * a valid chunk, else skip it.
   */
  fread(buf, 1, CHUNK_SIZE, zin);
  if ((buf[0]!=0xf5) && (buf[1]!=0x46) && (buf[3]!=0xbd)) parse_chunk(buf);
  else printf("Skipped useless chunk 0\n");

  while (fread(buf, 1, CHUNK_SIZE, zin) == CHUNK_SIZE) parse_chunk(buf);

  average= highest_total/highest_count;
  printf("Video - audio average #1 is %lu milliseconds\n",average);
  average= lowest_total/lowest_count;
  printf("Video - audio average #2 is %lu milliseconds\n",average);

  fclose(audio_file);
  fclose(video_file);
  exit(1);
}
